import asyncio
import os
from pathlib import Path
from typing import BinaryIO, Optional, TextIO, Union

from .AsyncTeeProtocol import AsyncTeeProtocol


async def async_exec(*argv: str,
                     files: Optional[dict[int, BinaryIO]] = None,
                     cwd: Optional[Path] = None,
                     env: Optional[dict[str, str]] = None) -> tuple[int, dict[int, bytearray]]:
    files = files if files is not None else {}
    loop = asyncio.get_running_loop()
    cmd_done = loop.create_future()
    proc = loop.subprocess_exec(lambda: AsyncTeeProtocol(cmd_done, files), *argv, stdin=None, cwd=cwd, env=env)
    transport = None
    try:
        transport, _ = await proc
        await cmd_done
    finally:
        if transport is not None:
            transport.close()

    return cmd_done.result()

def run(*argv: str,
        files: Optional[dict[Union[TextIO, BinaryIO], BinaryIO]] = None,
        cwd: Optional[Path] = None,
        env: Optional[dict[str, str]] = None) -> tuple[int, dict[int, bytearray]]:
    files = files if files is not None else {}
    env = env if env is not None else {}
    int_files = {fd.fileno(): f for fd, f in files.items()} # convert file objects to file descriptor
    return asyncio.run(async_exec(*argv, files=int_files, cwd=cwd, env={**os.environ, **env}))




from .Preprocessing import Preprocessing

class RemoveActivity(Preprocessing):
    def __init__(self, activities, noconvert: bool=False):
        from qualia_core.datamodel.har import Activities
        for i, activity in enumerate(activities):
            if not noconvert and isinstance(activity, str):
                activities[i] = Activities[activity]
        self.__activities = activities

    def __call__(self, datamodel):
        for _, subjectset in datamodel:
            for subject in subjectset:
                subject.activities = [activity for activity in subject.activities if activity.kind not in self.__activities]
        return datamodel



from dataclasses import dataclass, field

import numpy as np
import numpy.typing

from qualia_core.datamodel.sensor.Sensor import Sensor


@dataclass
class TimeSample:
    t: float
    sensors: list[Sensor] = field(default_factory=list)

    def get_raw_array(self) -> list[numpy.typing.NDArray[np.float32]]:
        return [v for s in self.sensors for v in s.get_raw_array()]



from __future__ import annotations

import sys
from pathlib import Path

from .Eclipse import Eclipse

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class STM32CubeIDE(Eclipse):
    def __init__(self,
                 projectname: str,
                 projectdir: Path,
                 outdir: Path | None = None) -> None:
        outdir = outdir if outdir is not None else Path('out')/'deploy'/'STM32CubeIDE'
        stm32cubeide_dir = Path('/opt')/'stm32cubeide'
        stm32cubeide_bin = stm32cubeide_dir/'stm32cubeide'

        # STM32CubeProgrammer bundled with STM32CubeIDE 1.6.0 cannot reset MCU, use system-wide STM32CubeProgrammer 2.7.0
        stm32cubeprogrammer_bin = Path('/')/'usr'/'bin'/'STM32_Programmer_CLI'
        #stm32cubeprogrammer_bin = next((self.__stm32cubeide_dir/'plugins').glob('com.st.stm32cube.ide.mcu.externaltools.cubeprogrammer.linux64_*'))/'tools'/'bin'/'STM32_Programmer.sh'

        arm_size_bin = next(filter(Path.is_dir, (stm32cubeide_dir/'plugins').glob('com.st.stm32cube.ide.mcu.externaltools.gnu-tools-for-stm32*')))/'tools'/'bin'/'arm-none-eabi-size'

        super().__init__(eclipse_bin=stm32cubeide_bin,
                         size_bin=arm_size_bin,
                         upload_bin=stm32cubeprogrammer_bin,
                         projectname=projectname,
                         projectdir=projectdir,
                         outdir=outdir)

    @override
    def _upload(self,
                tag: str,
                logdir: Path,
                args: tuple[str, ...] | None = None,
                cmd: Path | None = None) -> bool:
        args = ('--connect', 'port=SWD', 'mode=UR', 'reset=hwRst',
                '--download', str(self._outdir/f'{tag}.elf'), '0x08000000',
                '--verify',
                '-hardRst',
                *(args if args is not None else ()))
        return super()._upload(tag=tag, logdir=logdir, args=args, cmd=cmd)



from pathlib import Path
import re
import sys
import time
from qualia_core.evaluation.Stats import Stats
import numpy as np

class TFLite:
    def __init__(self, *args, **kwargs):
        pass

    def evaluate(self, framework, model_kind, dataset, target: str, tag: str, limit: int=None, dataaugmentations=[]):
        if dataaugmentations:
            raise ValueError(f'dataaugmentations not supported for {self.__class__.__name__}')

        import tensorflow as tf
        from sklearn.metrics import accuracy_score

        interpreter = tf.lite.Interpreter(model_path=str(Path('out')/'deploy'/'tflite'/f'{tag}.tflite'))
        interpreter.allocate_tensors()
        pred_test_q = np.ndarray(dataset.sets.test.y.shape)

        tstart = time.time() # Start timer
        for i, x in enumerate(dataset.sets.test.x):
            interpreter.set_tensor(interpreter.get_input_details()[0]['index'], [x])
            interpreter.invoke()
            pred_test_q[i] = interpreter.get_tensor(interpreter.get_output_details()[0]['index'])
        tstop = time.time() # Stop timer

        accuracy = accuracy_score(dataset.sets.test.y.argmax(axis=1), pred_test_q.argmax(axis=1))

        print('Quantized model accuracy: ', accuracy)
        print(tf.math.confusion_matrix(dataset.sets.test.y.argmax(axis=1), pred_test_q.argmax(axis=1)))

        return Stats(avg_time=(tstop - tstart) / len(dataset.sets.test.y), accuracy=accuracy)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from qualia_core.datamodel.DataModel import DataModel

T = TypeVar('T')

class Dataset(ABC, Generic[T]):
    sets: list[str]

    def __init__(self, sets: list[str] | None = None) -> None:
        super().__init__()
        self.sets = sets if sets is not None else list(DataModel.Sets.fieldnames())

    @abstractmethod
    def __call__(self) -> DataModel[T]:
        ...

    @abstractmethod
    def import_data(self) -> DataModel[T] | None:
        ...

    @property
    def name(self) -> str:
        return f'{self.__class__.__name__}'



from __future__ import annotations

import sys

import torch
import torch.nn
from torch import nn

from qualia_core.learningmodel.pytorch.Quantizer import QuantizationConfig, Quantizer, update_params

from .CustomBatchNorm import CustomBatchNorm
from .QuantizedLayer import QuantizedLayer, QuantizerActProtocol, QuantizerInputProtocol, QuantizerWProtocol

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override


class QuantizedLinear(nn.Linear, QuantizerInputProtocol, QuantizerActProtocol, QuantizerWProtocol, QuantizedLayer):
    bias: torch.nn.Parameter | None

    def __init__(self,  # noqa: PLR0913
                 in_features: int,
                 out_features: int,
                 quant_params: QuantizationConfig,
                 bias: bool = True,  # noqa: FBT001, FBT002
                 activation: nn.Module | None = None) -> None:
        self.call_super_init = True # Support multiple inheritance from nn.Module
        super().__init__(in_features, out_features, bias=bias)
        self.in_features = in_features
        self.out_features = out_features
        self.activation = activation
        # Create the quantizer instance
        quant_params_input = update_params(tensor_type = 'input', quant_params = quant_params)
        quant_params_act = update_params(tensor_type = 'act', quant_params = quant_params)
        quant_params_w = update_params(tensor_type = 'w', quant_params = quant_params)
        self.quantizer_input = Quantizer(**quant_params_input)
        self.quantizer_act = Quantizer(**quant_params_act)
        self.quantizer_w = Quantizer(**quant_params_w)
        if 'bias' in quant_params :
            quant_params_bias = update_params(tensor_type='bias', quant_params=quant_params)
            self.quantizer_bias = Quantizer(**quant_params_bias)


    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:  # noqa: A002
        #if self.quan_train and self.quand_test:

        q_input = self.quantizer_input(input)
        if self.bias is None:
            # If no bias, quantize weights only
            q_w = self.quantizer_w(self.weight)
            y = torch.nn.functional.linear(q_input, q_w)
        else :
            # Quantize bias and weights
            if hasattr(self, 'quantizer_bias'):
                #...with the different quantization schemes
                q_w = self.quantizer_w(self.weight)
                q_b = self.quantizer_bias(self.bias)
            else :
                #...with the same quantization schemes
                q_w, q_b = self.quantizer_w(self.weight, bias_tensor=self.bias)
            y = torch.nn.functional.linear(q_input, q_w, bias=q_b)

        if self.activation:
            y = self.activation(y)

        return self.quantizer_act(y)


class QuantizedReLU(torch.nn.ReLU, QuantizerInputProtocol, QuantizerActProtocol, QuantizedLayer):
    def __init__(self, quant_params: QuantizationConfig) -> None:
        self.call_super_init = True # Support multiple inheritance from nn.Module
        super().__init__()
        quant_params_input = update_params(tensor_type='input', quant_params=quant_params)
        quant_params_act = update_params(tensor_type='act', quant_params=quant_params)
        self.quantizer_input = Quantizer(**quant_params_input)
        self.quantizer_act = Quantizer(**quant_params_act)

    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:  # noqa: A002
        q_input = self.quantizer_input(input)

        y = super().forward(q_input)

        return self.quantizer_act(y)


class QuantizedIdentity(torch.nn.Identity, QuantizerInputProtocol, QuantizerActProtocol, QuantizedLayer):
    def __init__(self, quant_params: QuantizationConfig) -> None:
        self.call_super_init = True # Support multiple inheritance from nn.Module
        super().__init__()
        quant_params_act = update_params(tensor_type='act', quant_params=quant_params)
        self.quantizer_act = Quantizer(**quant_params_act)

        # Identity has same quantization on input and output
        # Only output quantizer used here but generation of activations_range.txt require input quantized info as well
        self.quantizer_input = self.quantizer_act

    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:  # noqa: A002

        y = super().forward(input)

        return self.quantizer_act(y)


class QuantizedBatchNorm(CustomBatchNorm, QuantizerInputProtocol, QuantizerActProtocol, QuantizerWProtocol, QuantizedLayer):
    def __init__(self,  # noqa: PLR0913
                 num_features: int,
                 quant_params: QuantizationConfig,
                 eps: float = 1e-05,
                 momentum: float = 0.1,
                 affine: bool = True,  # noqa: FBT001, FBT002
                 track_running_stats: bool = True,  # noqa: FBT002, FBT001
                 device: torch.device | None = None,
                 dtype: torch.dtype | None = None,
                 activation: torch.nn.Module | None = None) -> None:
        super().__init__(num_features=num_features,
                         eps=eps,
                         momentum=momentum,
                         affine=affine,
                         track_running_stats=track_running_stats,
                         use_fused_params=True,
                         device=device,
                         dtype=dtype)

        self.activation = activation

        # Create the quantizer instance
        quant_params_input = update_params(tensor_type='input', quant_params=quant_params)
        quant_params_act = update_params(tensor_type='act', quant_params=quant_params)
        quant_params_w = update_params(tensor_type='w', quant_params=quant_params)
        self.quantizer_input = Quantizer(**quant_params_input)
        self.quantizer_act = Quantizer(**quant_params_act)
        self.quantizer_w = Quantizer(**quant_params_w)
        if 'bias' in quant_params :
            quant_params_bias = update_params(tensor_type='bias', quant_params=quant_params)
            self.quantizer_bias = Quantizer(**quant_params_bias)

    @override
    def forward(self, input: torch.Tensor) -> torch.Tensor:  # noqa: A002

        input_shape = input.shape
        x = input.flatten(start_dim=2)
        q_input = self.quantizer_input(x)

        mean, var = self.compute_stats(q_input)

        if self.training and self.track_running_stats:
            self.update_running_stats(mean=mean, var=var)

        if self.training or not self.track_running_stats:
            alpha, beta = self.compute_fused_params(weight=self.weight,
                                                    bias=self.bias,
                                                    mean=mean,
                                                    var=var,
                                                    eps=self.eps)
        else:
            alpha, beta = self.compute_fused_params(weight=self.weight,
                                                    bias=self.bias,
                                                    mean=self.running_mean,
                                                    var=self.running_var,
                                                    eps=self.eps)

        # Quantize bias and weights
        if hasattr(self, 'quantizer_bias'):
            #...with the different quantization schemes
            q_alpha = self.quantizer_w(alpha)
            q_beta = self.quantizer_bias(beta)
        else :
            #...with the same quantization schemes
            q_alpha, q_beta = self.quantizer_w(alpha, bias_tensor=beta)

        y = self.compute_batchnorm_with_fused_params(x=q_input, alpha=q_alpha, beta=q_beta)

        if self.activation is not None:
            y = self.activation(y)

        return self.quantizer_act(y).reshape(input_shape)


# Keep imports there to avoid circular imports

from qualia_core.learningmodel.pytorch.quantized_layers1d import (  # noqa: E402
    QuantizedAdaptiveAvgPool1d,
    QuantizedAvgPool1d,
    QuantizedBatchNorm1d,
    QuantizedConv1d,
    QuantizedMaxPool1d,
)
from qualia_core.learningmodel.pytorch.quantized_layers2d import (  # noqa: E402
    QuantizedAdaptiveAvgPool2d,
    QuantizedAvgPool2d,
    QuantizedBatchNorm2d,
    QuantizedConv2d,
    QuantizedMaxPool2d,
)

from .QuantizedAdd import QuantizedAdd  # noqa: E402
from .QuantizedGlobalSumPool1d import QuantizedGlobalSumPool1d  # noqa: E402
from .QuantizedGlobalSumPool2d import QuantizedGlobalSumPool2d  # noqa: E402

quantized_layers = (
        QuantizedBatchNorm1d,
        QuantizedBatchNorm2d,
        QuantizedConv1d,
        QuantizedConv2d,
        QuantizedMaxPool1d,
        QuantizedMaxPool2d,
        QuantizedAvgPool1d,
        QuantizedAvgPool2d,
        QuantizedAdaptiveAvgPool1d,
        QuantizedAdaptiveAvgPool2d,
        QuantizedLinear,
        QuantizedAdd,
        QuantizedReLU,
        QuantizedIdentity,
        QuantizedGlobalSumPool1d,
        QuantizedGlobalSumPool2d,
        )



from __future__ import annotations

import logging
import sys
from typing import Any

logger = logging.getLogger(__name__)


class TensorFlowInitializer:
    __tensorflow_initialized = False

    def set_seed(self, seed: int) -> None:
        # From https://www.reddit.com/r/MachineLearning/comments/ge32bi/how_to_reliably_compare_experiments_when_tuning/fpl4owx/
        import os
        import sys

        os.environ['TF_DETERMINISTIC_OPS'] = '1'
        # If PYTHONHASHSEED is not set to 0 or CUBLAS_WORKSPACE_CONFIG not set to ':4096:8' (required for CuBlas determinism)
        # restart interpreter (env var has to be set before python starts)
        if os.environ.get('PYTHONHASHSEED', '') != str(seed) or os.environ.get('CUBLAS_WORKSPACE_CONFIG', '') != ':4096:8':
            os.execle(sys.executable,
                      sys.executable,
                      *sys.argv,
                      {'PYTHONHASHSEED': str(seed), 'CUBLAS_WORKSPACE_CONFIG': ':4096:8', **os.environ})

        import random
        random.seed(seed)

        import numpy as np
        np.random.seed(seed)  # noqa: NPY002 We do want to initialize the legacy global generator

        import qualia_core.random
        qualia_core.random.shared.seed(seed)

        try:
            import torch
            _ = torch.manual_seed(seed)  # type: ignore[untyped-def]

            import pytorch_lightning
            _ = pytorch_lightning.seed_everything(seed)
        except ImportError:
            logger.warning('PyTorch not loaded and not seeded')

        try:
            import tensorflow as tf  # type: ignore[import-untyped]
            tf.random.set_seed(seed)  # type: ignore[untyped-def]
        except ImportError:
            logger.warning('TensorFlow not loaded and not seeded')

    def __call__(self,
                 reserve_gpu: bool = True,  # noqa: FBT001, FBT002
                 gpu_memory_growth: bool = True,  # noqa: FBT001, FBT002
                 debug: bool = False,  # noqa: FBT001, FBT002
                 seed: int | None = None) -> None:
        if TensorFlowInitializer.__tensorflow_initialized:
            logger.warning('TensorFlow already initialized by TensorFlowInitialize, new settings not be applied properly.')
        elif 'tensorflow' in sys.modules:
            logger.error('TensorFlow already imported previously. TensorFlowInitializer must be called first!')
            raise RuntimeError

        if not debug:
            import os
            # Needs to be done before the first tensorflow import ever
            os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'  # or any {'0', '1', '2'}

        if seed is not None:
            self.set_seed(seed)

        try:
            import tensorflow as tf  # type: ignore[import-untyped]
        except ImportError:
            logger.warning('TensorFlow not loaded, not setting up devices')
            return

        if debug:
            tf.debugging.set_log_device_placement(enabled=True)  # type: ignore[untyped-def]

        if not reserve_gpu:
            # Hide GPUs when not training to not allocate any resource
            tf.config.set_visible_devices([], 'GPU')  # type: ignore[untyped-def]
        else:
            # From https://www.tensorflow.org/guide/gpu
            # Limiting GPU memory growth
            gpus: list[Any] = tf.config.experimental.list_physical_devices('GPU')  # type: ignore[untyped-def]
            # Currently, memory growth needs to be the same across GPUs
            for gpu in gpus:
                tf.config.experimental.set_memory_growth(gpu, gpu_memory_growth)  # type: ignore[untyped-def]

            logical_gpus: list[Any] = tf.config.experimental.list_logical_devices('GPU')  # type: ignore[untyped-def]
            logger.info('%d Physical GPUs, %d Logical GPUs', len(gpus), len(logical_gpus))

        TensorFlowInitializer.__tensorflow_initialized = True



from __future__ import annotations

import sys
from typing import Any, cast

import torch
from torch.autograd.function import Function, FunctionCtx

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

class IntNoGradientceil(Function):
    @staticmethod
    @override
    def forward(ctx: FunctionCtx, *args: torch.Tensor, **_: Any) -> torch.Tensor:
        return args[0].ceil()

    @staticmethod
    @override
    def backward(ctx: Function, *grad_outputs: torch.Tensor) -> tuple[torch.Tensor | None, None]:
        return grad_outputs[0], None

    @classmethod
    @override
    def apply(cls, *args: torch.Tensor, **kwargs: Any) -> torch.Tensor:
        return cast(torch.Tensor, super().apply(*args, **kwargs))  # type: ignore[no-untyped-call]

class IntNoGradientround(Function):
    @staticmethod
    @override
    def forward(ctx: FunctionCtx, *args: torch.Tensor, **_: Any) -> torch.Tensor:
        """Round to nearest.

        Half tie-breaker always rounds upwards to match Qualia-CodeGen's implementation,
        easier to implement than :meth:`torch.round`'s "round half to even" but seems to provide similar results.
        """
        return (args[0] + 0.5).floor()

    @staticmethod
    @override
    def backward(ctx: Function, *grad_outputs: torch.Tensor) -> tuple[torch.Tensor | None, None]:
        return grad_outputs[0], None

    @classmethod
    @override
    def apply(cls, *args: torch.Tensor, **kwargs: Any) -> torch.Tensor:
        return cast(torch.Tensor, super().apply(*args, **kwargs))  # type: ignore[no-untyped-call]

class IntNoGradientfloor(Function):
    @staticmethod
    @override
    def forward(ctx: FunctionCtx, *args: torch.Tensor, **_: Any) -> torch.Tensor:
        return args[0].floor()

    @staticmethod
    @override
    def backward(ctx: Function, *grad_outputs: torch.Tensor) -> tuple[torch.Tensor | None, None]:
        return grad_outputs[0], None

    @classmethod
    @override
    def apply(cls, *args: torch.Tensor, **kwargs: Any) -> torch.Tensor:
        return cast(torch.Tensor, super().apply(*args, **kwargs))  # type: ignore[no-untyped-call]
